// koggtag (C)2001 Dario Abatianni <eisfuchs@tigress.com>
// This program is distributed under the GNU General Public Licence (GPL)

#include <iostream>

#include <qpushbutton.h>
#include <qhbox.h>
#include <qvbox.h>
#include <qlabel.h>
#include <qlayout.h>
#include <qevent.h>
#include <qdragobject.h>

#include <kapp.h>
#include <kaction.h>
#include <kstdaction.h>
#include <klistview.h>
#include <klineedit.h>
#include <kfiledialog.h>
#include <kmessagebox.h>
#include <klocale.h>
#include <kconfig.h>
#include <kdialog.h>

#include "koggtagmainwindow.h"
#include "oggfile.h"

KOggTagMainWindow::KOggTagMainWindow(QString file="") : KMainWindow()
{
  /* Find out how big the margin between GUI components shall be */
  const int SPACING=KDialog::marginHint();
  /* Enable Drag & Drop events */
  setAcceptDrops(true);
  /* Add a box to the main window to hold the rest of the widgets */
  QVBox *box=new QVBox(this);
  /* Set the free space between the widgets */
  box->setSpacing(SPACING);
  /* Set the free space at the borders */
  box->setMargin(SPACING);

  /* Create an empty ogg file object */
  ogg_file=new OggFile();

  /* Add a box for the filename input field and label */
  QHBox *filenamebox=new QHBox(box);
  filenamebox->setSpacing(SPACING);
  /* Add "Ogg File" label */
  new QLabel(i18n("Ogg File"),filenamebox);
  /* Add file input field */
  filename_input=new KLineEdit(filenamebox);
  /* When the user presses Retun in the File Name field, the file should be loaded */
  connect(filename_input,SIGNAL (returnPressed()),this,SLOT (file_open_from_input()));

  /* Add Tag List View with two columns */
  taglist=new KListView(box);
  taglist->addColumn(i18n("Tag Name"));
  taglist->addColumn(i18n("Tag Value"));
  /* Let the columns automatically resize to maximum */
  taglist->setColumnWidthMode(0,QListView::Maximum);
  taglist->setColumnWidthMode(1,QListView::Maximum);
  /* Make this list a single selection list */
  taglist->setMultiSelection(false);
  taglist->setSelectionMode(QListView::Single);
  /* We want the whole line to become visibly selected */
  taglist->setAllColumnsShowFocus(true);
  /* Be sure to copy the newly selected item's name and value to the input fields */
  connect(taglist,SIGNAL (selectionChanged(QListViewItem *)),this,SLOT (doSelect(QListViewItem *)));
  /* Add a box for the Tag Name input field and label */
  QHBox *tagnamebox=new QHBox(box);
  tagnamebox->setSpacing(SPACING);
  new QLabel(i18n("Tag Name"),tagnamebox);
  tagname_input=new KLineEdit(tagnamebox);

  /* Add a box for the Tag Value input field and label */
  QHBox *tagvaluebox=new QHBox(box);
  tagvaluebox->setSpacing(SPACING);
  new QLabel(i18n("Tag Value"),tagvaluebox);
  tagvalue_input=new KLineEdit(tagvaluebox);

  /* Tag Name and Tag Value may become enabled only when a tag is selected */
  tagname_input->setDisabled(true);
  tagvalue_input->setDisabled(true);

  /* Update the selected item on changes in the input fields */
  connect(tagname_input,SIGNAL (textChanged(const QString &)),this,SLOT (updateItemTagName(const QString &)));
  connect(tagvalue_input,SIGNAL (textChanged(const QString &)),this,SLOT (updateItemTagValue(const QString &)));

  /* Add a box to the main box to hold the buttons */
  QHBox *buttonbox=new QHBox(box);
  buttonbox->setSpacing(SPACING);

  /* Add buttons to the box */
  addTagButton=new QPushButton(i18n("&Add Tag"),buttonbox);
  removeTagButton=new QPushButton(i18n("&Remove Tag"),buttonbox);
  QPushButton *closeButton=new QPushButton(i18n("&Close"),buttonbox);

  /* Disable the Add and Remove Tag buttons */
  addTagButton->setEnabled(false);
  removeTagButton->setEnabled(false);

  /* Add a new tag to the list when the Add Tag button is pressed */
  connect(addTagButton,SIGNAL (clicked()),this,SLOT (add_tag()));
  /* Remove the selected tag from the list when the Remove Tag button is pressed */
  connect(removeTagButton,SIGNAL (clicked()),this,SLOT (remove_tag()));
  /* Close window when "Close" button is pressed */
  connect(closeButton,SIGNAL (clicked()),this,SLOT (close()));

  /* Standard actions get added to the tool bars and menus automatically */
  KStdAction::openNew(this,SLOT(file_new()),actionCollection());
  KStdAction::open(this,SLOT(file_open()),actionCollection());
  saveAction=KStdAction::save(this,SLOT(file_save()),actionCollection());
  saveAsAction=KStdAction::saveAs(this,SLOT(file_save_as()),actionCollection());
  undoAction=KStdAction::undo(this,SLOT(file_redisplay()),actionCollection());
  KStdAction::close(this,SLOT(close()),actionCollection());

  /* The contents of this window have not been changed by the user yet */
  setChanged(false);
  /* Disable the Save As menu */
  saveAsAction->setEnabled(false);
  /* Make the outermost box the central widget so it gets layouted properly */
  setCentralWidget(box);
  /* KMainWindow::setCaption() automatically uses makeStdCaption() */
  setCaption("");

  createGUI();
  /* Was there a file name for us? */
  if(file.compare(""))
  {
    KURL url=file;
    /* Copy white space stripped file name into ogg_file object */
    ogg_file->setFullPath(url.path().stripWhiteSpace());
    ogg_file->setFilename(url.filename().stripWhiteSpace());
    /* Open this file */
    do_file_open();
  }
}

KOggTagMainWindow::~KOggTagMainWindow()
{
  /* Destroy the Ogg File object, since it is not used anymore */
  delete ogg_file;
}

bool KOggTagMainWindow::querySaveChanges()
{
  /* Was our file changed? */
  if(isChanged())
  {
    /* Generate Question string */
    QString question=QString(i18n("Some of your changes have not been saved yet!\n"
                                  "Do you want to save them now?"));
    /* Show a [Yes] [No] Cancel] Dialog */
    int result=KMessageBox::warningYesNoCancel(this,question,i18n("Unsaved Changes"));
    /* Check the result */
    if(result==KMessageBox::Cancel) return false;
    if(result==KMessageBox::Yes) file_save();
  }
  /* Return true as default */
  return true;
}
/* Create a new empty window */
void KOggTagMainWindow::file_new()
{
  KOggTagMainWindow *win=new KOggTagMainWindow();
  win->show();
}

void KOggTagMainWindow::file_open()
{
  /* Open File Dialog to get Ogg filename */
  const KURL url=KFileDialog::getOpenURL("","*.ogg");
  /* Check, if User entered a filename odr hit cancel */
  if(url.isEmpty()==false)
  {
    /* Check if filename is valid */
    if(url.isMalformed()==false)
    {
      /* Check if this window already has an open file */
      if(ogg_file->isOpen())
      {
        /* Open a new window with this path */
        KOggTagMainWindow *win=new KOggTagMainWindow(url.path());
        /* Show the new window */
        win->show();
      }
      else
      {
        /* Set the Ogg Filename */
        ogg_file->setFullPath(url);
        /* Open this file */
        do_file_open();
      }
    }
  }
}
/* This method opens the file specified in the File name Input */
void KOggTagMainWindow::file_open_from_input()
{
  /* returnPressed()-Signal seems to be sent twice, so we check if the name
     actually has changed before opening the file */
  if(ogg_file->getFullPath().compare(filename_input->text()))
  {
    /* Only try to load the file of the input field is not empty */
    if(filename_input->text().compare(""))
    {
      /* Ask the user if they want to save changes */
      if(querySaveChanges())
      {
        /* Copy file name into ogg_file object */
        ogg_file->setFullPath(filename_input->text());
        /* Open this file */
        do_file_open();
      }
    }
  }
  else cerr << "FIXME: Tried to open the same file again. Ignored." << endl;
}
/* This method opens the file specified in ogg_file->getFullPath() */
void KOggTagMainWindow::do_file_open()
{
  if(ogg_file->isOpen())
  {
    ogg_file->unload();
    taglist->clear();
    /* Disable the Remove Tag button */
    removeTagButton->setDisabled(true);
    /* Tag Name and Tag Value may become enabled only when a tag is selected */
    tagname_input->setDisabled(true);
    tagvalue_input->setDisabled(true);
    /* Disable the Save and Save As buttons */
    saveAction->setEnabled(false);
    saveAsAction->setEnabled(false);
  }
  /* Let the Ogg File open it's data */
  ogg_file->load();
  /* load() succeded? */
  if(ogg_file->isOpen())
  {
    Tag *tag=ogg_file->tags->first();
    while(tag)
    {
      add_tag(tag->getName(),tag->getValue());
      tag=ogg_file->tags->next();
    }
    /* Enable the Add Tag button */
    addTagButton->setEnabled(true);
    /* Enable the Save As menu */
    saveAsAction->setEnabled(true);
    /* Update the filename box */
    filename_input->setText(ogg_file->getFullPath());
  }
  else
  {
    /* Display error dialog */
    QString error=(i18n("Could not load file:\n\"%1\"").arg(ogg_file->getFullPath()));
    KMessageBox::sorry(this,error,i18n("Error opening File"));
    /* Clear file input widget */
    filename_input->setText("");
  }
  /* Freshly loaded or no loaded files have no changes */
  setChanged(false);
}

void KOggTagMainWindow::file_save()
{
  /* Write out the Ogg File data with its associated filename (NULL) */
  file_save(NULL);
}

void KOggTagMainWindow::file_save(QString *new_file_name)
{
  if(ogg_file->isOpen())
  {
    /* Delete all tags from the Ogg File */
    ogg_file->tags->clear();

    /* Traverse through the Tag List and add them to the Ogg File */
    QListViewItem *tag=taglist->firstChild();
    while(tag)
    {
      ogg_file->tags->append(new Tag(tag->text(0),tag->text(1)));
      tag=tag->nextSibling();
    }
    /* Tell the Ogg File to write its data (if the name is present it will "Save As") */
    if(ogg_file->write(new_file_name))
    {
      /* Freshly saved files have no changes */
      setChanged(false);
    }
    else
    {
      /* Display error dialog */
      QString error=(i18n("Could not save file:\n\"%1\"").arg(ogg_file->getFullPath()));
      KMessageBox::sorry(this,error,i18n("Error saving File"));
      /* Clear file input widget */
      filename_input->setText("");
    }
  }
}

void KOggTagMainWindow::file_save_as()
{
  /* Open File Dialog to get Ogg filename */
  KURL filename=KFileDialog::getOpenURL(ogg_file->getFullPath(),"*.ogg");
  /* Check, if User entered a filename odr hit cancel */
  if(filename.isEmpty()==false)
  {
    /* Check if filename is valid */
    if(filename.isMalformed()==false)
    {
      /* Get filename for new file */
      QString newname=filename.path();
      /* Save the tags into this file */
      file_save(&newname);
      /* Set the new Ogg File Path */
      ogg_file->setFullPath(filename.path());
      /* Update the filename box */
      filename_input->setText(ogg_file->getFullPath());
    }
  }
}

void KOggTagMainWindow::file_redisplay()
{
  /* Reopen the file, if one was opened already */
  if(ogg_file->isOpen()) do_file_open();
}
/* This method gets called whenever one of the windows should be closed */
bool KOggTagMainWindow::queryClose()
{
  /* Ask the user if they want to save changes and return the value */
  return querySaveChanges();
}

void KOggTagMainWindow::add_tag()
{

  /* Insert a new item into the list (should be at the last position, not sorted somewhere) */
  KListViewItem *item=new KListViewItem(taglist,"","");

  /* Remember that the user changed something */
  setChanged(true);
  /* Take care of all things to select an item */
  doSelect(item);
}

void KOggTagMainWindow::add_tag(QString& new_name,QString& new_value)
{
  /* Add new tag to the list */
  new KListViewItem(taglist,new_name,new_value);
}

void KOggTagMainWindow::remove_tag()
{
  /* Find out which item was selected */
  QListViewItem *selected_item=taglist->currentItem();

  /* Was there really a selected item? */
  if(selected_item)
  {
    /* Find the items above and below the selected one */
    QListViewItem *below=selected_item->itemBelow();
    QListViewItem *above=selected_item->itemAbove();

    /* If there were no more items, remove the selection altogether */
    if(below==0 && above==0) doSelect(NULL);
    /* Select the item below the old item, or above it, if there was none */
    else doSelect((below) ? below : above);

    /* Remove the item */
    delete selected_item;
    /* Remember that the user changed something */
    setChanged(true);
  }
}

void KOggTagMainWindow::doSelect(QListViewItem *selected_item)
{
  /* Is there a selected item? */
  if(selected_item)
  {
    /* Enable the tag name and value input fileds */
    tagname_input->setEnabled(true);
    tagvalue_input->setEnabled(true);

    /* Select the new tag, deselecting all others. */
    taglist->setSelected(selected_item,true);
    /* Enable the Remove Tag button */
    removeTagButton->setEnabled(true);
    /* Copy the text of the selected item into the input fields */
    tagname_input->setText(selected_item->text(0));
    tagvalue_input->setText(selected_item->text(1));
    taglist->ensureItemVisible(selected_item);
  }
  else /* No item selected */
  {
    /* Disable the tag name and value input fileds */
    tagname_input->setEnabled(false);
    tagvalue_input->setEnabled(false);
    /* Disable the Remove Tag button */
    removeTagButton->setEnabled(false);
    /* Clear the text of the input fields */
    tagname_input->setText("");
    tagvalue_input->setText("");
  }
}

void KOggTagMainWindow::updateItemTagName(const QString & string)
{
  /* Drag & Drop into a disabled Text Field would lead to a segfault */
  if(ogg_file->isOpen())
  {
    /* Find out which item is currently selected (check if there was none) */
    QListViewItem *selected_item=taglist->currentItem();
    /* Did the user change the text or was this an update after selecting another item? */
    if(selected_item && string!=selected_item->text(0))
    {
      /* Update the text of the name column (0) with the given string */
      selected_item->setText(0,string);
      /* Remember that the user changed something */
      setChanged(true);
    }
  }
}

void KOggTagMainWindow::updateItemTagValue(const QString & string)
{
  /* Drag & Drop into a disabled Text Field would lead to a segfault */
  if(ogg_file->isOpen())
  {
    /* Find out which item is currently selected (check if there was none) */
    QListViewItem *selected_item=taglist->currentItem();
    /* Did the user change the text or was this an update after selecting another item? */
    if(selected_item && string!=selected_item->text(1))
    {
      /* Update the text of the value column (1) with the given string */
      selected_item->setText(1,string);
      /* Remember that the user changed something */
      setChanged(true);
    }
  }
}

/* This method will be called whenever an icon is dragged over the window */
void KOggTagMainWindow::dragEnterEvent(QDragEnterEvent *drag)
{
  /* Accept all drops that can be translated into text/plain (file path) */
  if(drag->provides("text/plain")) drag->accept(rect());
  else drag->ignore(rect());
}

/* This method will be called when a dragged file is being dropped onto the window */
void KOggTagMainWindow::dropEvent(QDropEvent *drop)
{
  QString name;
  /* Decode the data into file path */
  if(QTextDrag::decode(drop,name))
  {
    int pos;
    /* As long as there is a file name left in the drop list ... */
    while((pos=name.find("\n"))!=-1)
    {
      /* Check if this window already has an opened file */
      if(ogg_file->isOpen())
      {
        /* Create a new window with the entry up to the \n (will be
           white space stripped there) */
        KOggTagMainWindow *win=new KOggTagMainWindow(name.left(pos));
        /* Show the window */
        win->show();
      }
      else
      {
        /* Get the entry up to the \n */
        KURL url=name.left(pos);
        /* Copy white space stripped file name into ogg_file object */
        ogg_file->setFullPath(url.path().stripWhiteSpace());
        ogg_file->setFilename(url.filename().stripWhiteSpace());
        /* Open this file */
        do_file_open();
      }
      /* Cut the name from the list and move on to the next entry*/
      name=name.mid(pos+1);
    } /* while */
    /* Tell the souce we accepted the file */
    drop->accept();
  }
  else cerr << "KOggTagMainWindow::dropEvent(): Could not decode data!" << endl;
}

void KOggTagMainWindow::setChanged(bool state)
{
  /* Set the changed flag according to the parameter */
  is_changed=state;
  undoAction->setEnabled(state);
  saveAction->setEnabled(state);
  /* Update the application caption with file name and maybe modified indicator */
  /* KMainWindow::setCaption() automatically uses makeStdCaption() */
  if(ogg_file->isOpen())
    setCaption(ogg_file->getFilename(),state);
  else
    setCaption("");
}

bool KOggTagMainWindow::isChanged()
{
  return is_changed;
}
