// koggtag (C)2001 Dario Abatianni <eisfuchs@tigress.com>
// This program is distributed under the GNU General Public Licence (GPL)

// #include <stdio.h>

#include <iostream>

#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>

#include <qstring.h>

#include <vorbis/vorbisfile.h>

#include "oggfile.h"
#include "tag.h"
#include "vcedit.h"

OggFile::OggFile()
{
  /* Create new tag objects */
  tags=new QList<Tag>();

  /* There was no data loaded yet */
  setOpen(false);
}

OggFile::~OggFile()
{
  /* Destroy the tag list object */
  delete tags;
}

void OggFile::load()
{
  OggVorbis_File file;

  int success=0;

  FILE *handle=fopen(getFullPath().latin1(),"r");
  if(handle==0) cerr << "OggFile::load(): fopen('" << getFullPath().latin1() << "')==0!" << endl;
  else
  {
    if(ov_open(handle,&file,NULL,0))
    {
      cerr << "OggFile::load(): ov_open('" << getFullPath().latin1() << "')!=0!" << endl;

      fclose(handle); /* only in case of failure. otherwise use ov_clear() to close! */
    }
    else
    {
      struct vorbis_comment *comment;

      comment=ov_comment(&file,-1);
      if(comment==0) cerr << "OggFile::load(): ov_comment()==0!" << endl;
      else
      {
        int tagnum=0;
        char **taglist=comment->user_comments;

        while(taglist[tagnum])
        {
          tags->append(new Tag(taglist[tagnum]) );
          tagnum++;
        }
        success=1;
      }
      ov_clear(&file); /* ogg vorbis library fclose()s the file for us */
    }
  }

  if(!success) unload();
  setOpen(success);
}

void OggFile::unload()
{
  /* Remove all tags from Ogg File */
  while(tags->removeFirst());
  /* Declare file as closed, i.e. no data loaded */
  setOpen(false);
}

void OggFile::load(const KURL& openname)
{
  /* Set the filename for this Ogg File */
  setFullPath(openname);
  /* Open it */
  load();
}

void OggFile::setOpen(bool open)
{
  /* Set the open flag according to the parameter */
  is_open=open;
}

bool OggFile::isOpen()
{
  return is_open;
}

bool OggFile::write(QString *new_file_name)
{
  bool dorename=false;
  char tmpname[14];

  FILE *handle=fopen(getFullPath().latin1(),"r");
  if(handle==0) cerr << "OggFile::write(): fopen('" << getFullPath().latin1() << "')==0!" << endl;
  else
  {
    vcedit_state *state=vcedit_new_state();
    if(vcedit_open(state,handle)==-1) cerr << "OggFile::write(): vcedit_open('" << getFullPath().latin1() << "')==-1!" << endl;
    else
    {
      QString *vendor=new QString();

      struct vorbis_comment *oc=vcedit_comments(state);
      if(oc)
      {
        vendor->setLatin1(oc->vendor);
      }

      struct vorbis_comment *vc=state->vc;
      if(vc) vorbis_comment_clear(vc);

      vc->vendor=(char *)calloc(1,vendor->length()+1);
      strcpy(vc->vendor,vendor->latin1());

      Tag *node=tags->first();
      while(node)
      {
        char *name;
        char *value;

        name=(char *) calloc(1,node->getName().length()+1);
        value=(char *) calloc(1,node->getValue().length()+1);

        strcpy(name,node->getName().latin1());
        strcpy(value,node->getValue().latin1());

        vorbis_comment_add_tag(vc,name,value);

        free(value);
        free(name);

        node=tags->next();
      }

      strcpy(tmpname,"koggfileXXXXXX");

      int tmphandle=mkstemp(tmpname);
      if(tmphandle==-1) cerr << "OggFile::write(): tmphandle==-1!" << endl;
      else
      {
        close(tmphandle);

        FILE *outfile=fopen(tmpname,"w");
        if(outfile==0)
        {
          cerr << "OggFile::write(): outfile==0!" << endl;
          vcedit_clear(state); // frees comment entries and vendor
        }
        else
        {
          vcedit_write(state,outfile); // calls vcedit_clear() itself so we don't free anything
        }

        fclose(outfile);
        dorename=1;
      }

      free(vc->vendor);
    }
    fclose(handle);
  }
  /* If everything went fine rename the new ogg file to the old name */
  if(dorename)
  {
    int success;
    /* Check if renaming the file succeeded */
    success=rename(tmpname,(new_file_name) ? new_file_name->latin1() : getFullPath().latin1() );
    /* If not, try to copy the file instead */
    if(success==-1)
    {
      int inhandle;
      /* Open the temporary source file */
      inhandle=open(tmpname,O_RDONLY);
      /* Check if it worked */
      if(inhandle==-1) dorename=false;
      else
      {
        int outhandle,len;
        char buffer[1024];
        /* Try to create and open the final file */
        outhandle=open((new_file_name) ? new_file_name->latin1() : getFullPath().latin1(),O_WRONLY | O_CREAT,0644);
        /* Reset flag if this not succeed */
        if(outhandle==-1) dorename=false;
        else
        {
          /* Read the file in 1 KB chunks and write it out into the final file */
          do
          {
            len=read(inhandle,buffer,1024);
            std::write(outhandle,buffer,len);
          } while(len==1024);
          /* Close the final file */
          close(outhandle);
        }
        /* Close the source file */
        close(inhandle);
      }
    }
  }
  /* Delete the temporary file that may be still there */
  unlink(tmpname);
  /* Return success flag */
  return dorename;
}

const QString& OggFile::getFilename()
{
  return filename;
}

const QString& OggFile::getFullPath()
{
  return path;
}

void OggFile::setFullPath(const KURL& newpath)
{
  /* Set the new path and file name of this Ogg File */
  path=newpath.path();
  filename=newpath.fileName();
}

void OggFile::setFullPath(const QString& newpath)
{
  /* Set the new path and file name of this Ogg File */
  path=newpath;
  filename=newpath;
}

void OggFile::setFilename(const char *newname)
{
  /* Set the new name for the file */
  filename=newname;
}
